import 'dart:async';

import 'package:api/converter/request_response/payment/driven_distance_distributed.dart';
import 'package:api/model/interface/payment/driven_distance_distributed.dart';
import 'package:api/model/request/payment/driven_distance_distributed.dart';
import 'package:api/model/response/payment/driven_distance_distributed.dart';
import 'package:api/model/response/route.dart';
import 'package:api/model/team.dart';
import 'package:api/model/user.dart';
import 'package:lib/lib.dart';

class DrivenDistanceDistributedPaymentApiEndpoint extends CheckedDbApiEndpoint<DrivenDistanceDistributedPaymentRequest, DrivenDistanceDistributedPaymentResponse> {
  final User requestingUser;
  final Iterable<Team> requestingUserTeams;
  final Future<RouteResponse> Function(String) getRouteFromId;

  @override
  Future<bool> Function(DrivenDistanceDistributedPaymentRequest) get createItemAllowed => _createItemAllowed;

  @override
  Future<bool> Function(DrivenDistanceDistributedPaymentResponse) get getItemAllowed => _getItemAllowed;

  @override
  Future<bool> Function(DrivenDistanceDistributedPaymentRequest, DrivenDistanceDistributedPaymentResponse) get updateItemAllowed => _updateItemAllowed;

  @override
  Future<bool> Function(DrivenDistanceDistributedPaymentResponse) get deleteItemAllowed => _deleteItemAllowed;
  
  DrivenDistanceDistributedPaymentApiEndpoint({
    required super.database,
    required this.requestingUser,
    required this.requestingUserTeams,
    required this.getRouteFromId,
    super.itemFromRequest = DrivenDistanceDistributedPaymentRequestResponseConverter.fromRequest,
    super.itemToResponse = DrivenDistanceDistributedPaymentRequestResponseConverter.toResponseData,
  });

  Future<bool> _createItemAllowed(DrivenDistanceDistributedPaymentRequest item) async {
    return _isUserOwner(item)
      && _isUserInTeam(item)
      && _hasAtLeastOneRoute(item)
      && await _areRoutesInSameTeam(item)
      && _isAmountOk(item);
  }

  Future<bool> _getItemAllowed(DrivenDistanceDistributedPaymentResponse item) {
    return Future.value(
      _isUserInTeam(item)
    );
  }

  Future<bool> _updateItemAllowed(DrivenDistanceDistributedPaymentRequest item, DrivenDistanceDistributedPaymentResponse originalItem) async {
    return item.id == originalItem.id
      && _isUserOwner(item)
      && _isUserOwner(originalItem)
      && _isUserInTeam(item)
      && _hasAtLeastOneRoute(item)
      && await _areRoutesInSameTeam(item)
      && _isAmountOk(item);
  }

  Future<bool> _deleteItemAllowed(DrivenDistanceDistributedPaymentResponse item) {
    return Future.value(
      _isUserOwner(item)
    );
  }

  bool _isUserOwner(DrivenDistanceDistributedPayment item) {
    return item.paidBy == requestingUser.id;
  }

  bool _isUserInTeam(DrivenDistanceDistributedPayment item) {
    return requestingUserTeams.any((e) => e.id == item.id);
  }

  bool _hasAtLeastOneRoute(DrivenDistanceDistributedPayment item) {
    return item.routeIds.isNotEmpty;
  }

  Future<bool> _areRoutesInSameTeam(DrivenDistanceDistributedPayment item) async {
    if(item.routeIds.isEmpty) {
      return true;
    }
    var routes = Stream.fromFutures(item.routeIds.map((e) => getRouteFromId(e)));

    var teamToBeComparedTo = await routes.first;
    return routes.every((e) => e.teamId == teamToBeComparedTo.teamId);
  }

  bool _isAmountOk(DrivenDistanceDistributedPayment item) {
    return item.amount >= 0;
  }
}
