import 'dart:async';
import 'dart:io';
import 'package:api/api_endpoint/calendar_event.dart';
import 'package:api/api_endpoint/membership.dart';
import 'package:api/api_endpoint/payment/driven_distance_distributed.dart';
import 'package:api/api_endpoint/payment/equally_distributed.dart';
import 'package:api/api_endpoint/payment/user_to_user.dart';
import 'package:api/database/calendar_event.dart';
import 'package:api/api_endpoint/debt.dart';
import 'package:api/model/response/payment/driven_distance_distributed.dart';
import 'package:api/model/response/payment/equally_distributed.dart';
import 'package:api/model/response/payment/user_to_user.dart';
import 'package:api/model/team.dart';
import 'package:api/model/user.dart';
import 'package:api/service/debt.dart';
import 'package:api/database/payment/driven_distance_distributed.dart';
import 'package:api/database/payment/equally_distributed.dart';
import 'package:api/database/payment/user_to_user.dart';
import 'package:api/api_endpoint/team.dart';
import 'package:api/service/distance_calculator.dart';
import 'package:api/service/membership_service.dart';
import 'package:api/service/team_service.dart';
import 'package:lib/lib.dart';
import 'package:api/api_endpoint/point.dart';
import 'package:api/database/point.dart';
import 'package:api/api_endpoint/route.dart';
import 'package:api/database/route.dart';
import 'package:api/api_endpoint/schedule.dart';
import 'package:api/api_endpoint/user.dart';
import 'package:api/service/user_service.dart';

const String scheduleApiPath = "/";
const String pointApiPath = "/point";
const String routeApiPath = "/route";
const String calendarEventApiPath = "/calendar";
const String teamApiEndpoint = "/team";
const String userApiPath = "/user";
const String drivenDistanceDistributedPaymentApiEndpoint = "/payment/team/drivendistance";
const String equallyDistributedPaymentApiEndpoint = "/payment/team/equal";
const String userToUserPaymentApiEndpoint = "/payment/user";
const String debtApiEndpoint = "/debt";
const String membershipApiEndpoint = "/membership";

Future<dynamic> main(final context) async {
  final host = "https://${context.req.host}/v1";
  final apiKey = Platform.environment['FUNCTION_API_KEY'] ?? "";
  final projectId = AppwriteVariables.projectId;
  final String requestPath = context.req.path;

  var pointDatabase = PointDatabase(
    host: host,
    projectId: projectId,
    apiKey: apiKey,
  );

  var routeDatabase = RouteDatabase(
    host: host,
    projectId: projectId,
    apiKey: apiKey,
  );

  var calendarEventDatabase = CalendarEventDatabase(
    host: host,
    projectId: projectId,
    apiKey: apiKey,
  );

  var drivenDistanceDistributedPaymentDatabase = DrivenDistanceDistributedPaymentDatabase(
    host: host,
    projectId:
    projectId,
    apiKey: apiKey,
  );

  var equallyDistributedPaymentDatabase = EquallyDistributedPaymentDatabase(
    host: host,
    projectId: projectId,
    apiKey: apiKey,
  );

  var userToUserPaymentDatabase = UserToUserPaymentDatabase(
    host: host,
    projectId: projectId,
    apiKey: apiKey,
  );

  var teamService = TeamService(host: host, projectId: projectId, apiKey: apiKey);
  var userService = UserService(host: host, projectId: projectId, apiKey: apiKey);
  var membershipService = MembershipService(host: host, projectId: projectId, apiKey: apiKey, teamService: teamService, userService: userService);
  var distanceService = DistanceService(getPointById: pointDatabase.get);

  var debtService = DebtService(
    getMembershipsByTeamId: membershipService.getTeamMemberships,
    getRouteById: routeDatabase.get,
    getDistance: distanceService.calculateDistance,
    getUserToUserPayments: (userId) => getUserToUserPayments(userId, userToUserPaymentDatabase),
    getEquallyDistributedPayments: (userId) => getEquallyDistributedPayments(userId, membershipService, equallyDistributedPaymentDatabase),
    getDrivenDistanceDistributedPayments: (userId) => getDrivenDistanceDistributedPayments(userId, membershipService, drivenDistanceDistributedPaymentDatabase),
  );

  ApiEndpoint apiEndpoint;
  switch (requestPath) {
    case scheduleApiPath:
      apiEndpoint = ScheduleApiEndpoint(trigger: AppwriteVariables.trigger(context));
      break;

    case pointApiPath:
      apiEndpoint = PointApiEndpoint(
        database: pointDatabase,
      );
      break;

    case routeApiPath:
      var requestingUser = await getRequestingUser(context, userService);
      var requestingUserTeams = await getUserTeams(requestingUser.id, membershipService).toList();

      apiEndpoint = RouteApiEndpoint(
        database: routeDatabase,
        requestingUser: requestingUser,
        requestingUserTeams: requestingUserTeams,
      );
      break;

    case teamApiEndpoint:
      var requestingUser = await getRequestingUser(context, userService);

      apiEndpoint = TeamApiEndpoint(
        requestingUser: requestingUser,
        getTeamById: teamService.getTeam,
        isUserInTeam: (user, team) => isUserInTeam(user, team, membershipService),
      );
      break;

    case calendarEventApiPath:
      var requestingUser = await getRequestingUser(context, userService);
      var requestingUserTeams = await getUserTeams(requestingUser.id, membershipService).toList();

      apiEndpoint = CalendarEventApiEndpoint(
        database: calendarEventDatabase,
        requestingUser: requestingUser,
        requestingUserTeams: requestingUserTeams,
      );
      break;

    case userApiPath:
      var requestingUser = await getRequestingUser(context, userService);
      apiEndpoint = UserApiEndpoint(
        requestingUser: requestingUser,
        getTeamsOfUser: (user) => getUserTeams(user.id, membershipService),
        getUserFromId: userService.getUser,
      );
      break;

    case drivenDistanceDistributedPaymentApiEndpoint:
      var requestingUser = await getRequestingUser(context, userService);
      apiEndpoint = DrivenDistanceDistributedPaymentApiEndpoint(
        database: drivenDistanceDistributedPaymentDatabase,
        requestingUser: requestingUser,
        requestingUserTeams: await getUserTeams(requestingUser.id, membershipService).toList(),
        getRouteFromId: routeDatabase.get,
      );
      break;

    case equallyDistributedPaymentApiEndpoint:
      var requestingUser = await getRequestingUser(context, userService);
      apiEndpoint = EquallyDistributedApiEndpoint(
        database: equallyDistributedPaymentDatabase,
        requestingUser: requestingUser,
        requestingUserTeams: await getUserTeams(requestingUser.id, membershipService).toList(),
      );
      break;

    case userToUserPaymentApiEndpoint:
      var requestingUser = await getRequestingUser(context, userService);
      apiEndpoint = UserToUserPaymentApiEndpoint(
        database: userToUserPaymentDatabase,
        requestingUser: requestingUser,
      );
      break;
    
    case debtApiEndpoint:
      var requestingUser = await getRequestingUser(context, userService);
      apiEndpoint = DebtApiEndpoint(
        debtService: debtService,
        requestingUser: requestingUser,
      );
      break;

    case membershipApiEndpoint:
      var requestingUser = await getRequestingUser(context, userService);
      apiEndpoint = MembershipApiEndpoint(
        requestingUser: requestingUser,
        getMembershipsByTeamId: membershipService.getTeamMemberships,
        getMembershipsByUserId: membershipService.getUserMemberships,
      );

    default:
      throw HttpException("Path $requestPath not found", uri: Uri.parse(context.req.url));
  }

  var requestHandler = RequestHandler(apiEndpoint: apiEndpoint);
  return requestHandler.handleRequest(context);
}

Future<User> getRequestingUser(final context, UserService userService) async {
  final userId = AppwriteVariables.userId(context);
  return userService.getUser(userId);
}

Stream<Team> getUserTeams(String userId, MembershipService membershipService) {
  return membershipService.getUserMemberships(userId).map((e) => e.team);
}

Stream<UserToUserPaymentResponse> getUserToUserPayments(String userId, UserToUserPaymentDatabase userToUserPaymentDatabase) {
  return userToUserPaymentDatabase.getAll().where((e) => e.paidBy == userId || e.paidTo == userId);
}

Stream<EquallyDistributedPaymentResponse> getEquallyDistributedPayments(String userId, MembershipService membershipService, EquallyDistributedPaymentDatabase equallyDistributedPaymentDatabase) async* {
  var userTeamIds = await membershipService.getUserMemberships(userId).map((e) => e.team.id).toList();
  yield* equallyDistributedPaymentDatabase.getAll().where((e) => userTeamIds.contains(e.teamId)); // Would be nice with queries, but appwrite does not support nested queries
}

Stream<DrivenDistanceDistributedPaymentResponse> getDrivenDistanceDistributedPayments(String userId, MembershipService membershipService, DrivenDistanceDistributedPaymentDatabase drivenDistanceDistributedPaymentDatabase) async* {
  var userTeamIds = await membershipService.getUserMemberships(userId).map((e) => e.team.id).toList();
  yield* drivenDistanceDistributedPaymentDatabase.getAll().where((e) => userTeamIds.contains(e.teamId)); // Would be nice with queries, but appwrite does not support nested queries
}

Future<bool> isUserInTeam(User user, Team team, MembershipService membershipService) {
  return membershipService.getUserMemberships(user.id).any((e) => e.team.id == team.id);
}
