import 'package:api/model/debt.dart';
import 'package:api/model/interface/payment/user_to_user.dart';
import 'package:api/model/interface/route.dart';
import 'package:api/model/membership.dart';
import 'package:api/model/interface/payment/driven_distance_distributed.dart';
import 'package:api/model/response/payment/equally_distributed.dart';

class DebtService {
  final Stream<Membership> Function(String) getMembershipsByTeamId;
  final Future<Route> Function(String) getRouteById;
  final Stream<UserToUserPayment> Function(String) getUserToUserPayments;
  final Stream<EquallyDistributedPaymentResponse> Function(String) getEquallyDistributedPayments;
  final Stream<DrivenDistanceDistributedPayment> Function(String) getDrivenDistanceDistributedPayments;
  final Future<num> Function(Route) getDistance;

  DebtService({
    required this.getMembershipsByTeamId,
    required this.getRouteById,
    required this.getUserToUserPayments,
    required this.getEquallyDistributedPayments,
    required this.getDrivenDistanceDistributedPayments,
    required this.getDistance,
  });

  Stream<Debt> getDebt(String userId) async* {
    var debt = <Debt>[];

    await for (var payment in getUserToUserPayments(userId)) {
      debt.add(Debt(creditor: payment.paidBy, debtor: payment.paidTo, amount: payment.amount));
    }
    debt = await _reduce(debt, userId).toList();

    await for (var payment in getEquallyDistributedPayments(userId)) {
      debt.addAll(await calculateEquallyDistributedPaymentDebt(payment).toList());
    }
    debt = await _reduce(debt, userId).toList();

    await for (var payment in getDrivenDistanceDistributedPayments(userId)) {
      debt.addAll(await calculateDrivenDistanceDistributedPaymentDebt(payment).toList());
    }
  }

  Stream<Debt> calculateEquallyDistributedPaymentDebt(EquallyDistributedPaymentResponse payment) async* {
    var teamMemberships = getMembershipsByTeamId(payment.teamId);
    var teamMemberIds = await teamMemberships.where((element) => element.joinedAt.compareTo(payment.createdAt) <= 0).map((e) => e.user.id).toList();

    var debtAmountPerUser = payment.amount / teamMemberIds.length;

    for (var teamMemberId in teamMemberIds) {
      if (teamMemberId != payment.paidBy) {
        yield Debt(creditor: payment.paidBy, debtor: teamMemberId, amount: debtAmountPerUser);
      }
    }
  }

  Stream<Debt> calculateDrivenDistanceDistributedPaymentDebt(DrivenDistanceDistributedPayment payment) async* {
    var paymentRoutes = Stream.fromFutures(payment.routeIds.map((e) => getRouteById(e)));
    var distancePerUser = <String, num>{};

    await for (var route in paymentRoutes) {
      distancePerUser[route.userId] = (distancePerUser[route.userId] ?? 0) + await getDistance(route);
    }

    var fullDistance = distancePerUser.values.reduce((value, element) => value + element);

    for (var userDistance in distancePerUser.entries) {
      if (userDistance.key != payment.paidBy) {
        var debtAmount = payment.amount * userDistance.value / fullDistance;
        yield Debt(creditor: payment.paidBy, debtor: userDistance.key, amount: debtAmount);
      }
    }
  }

  Stream<Debt> _reduce(Iterable<Debt> debt, String userId) async* {
    var combinedDebt = <String, num>{};

    for (var debtEntries in debt.where((e) => e.creditor == userId)) {
      var summedDebtAmount = (combinedDebt[debtEntries.debtor] ?? 0) - debtEntries.amount;
      combinedDebt[debtEntries.debtor] = summedDebtAmount;
    }

    for (var debtEntries in debt.where((e) => e.debtor == userId)) {
      var summedDebtAmount = (combinedDebt[debtEntries.creditor] ?? 0) + debtEntries.amount;
      combinedDebt[debtEntries.creditor] = summedDebtAmount;
    }

    var combinedDebtEntries = combinedDebt.entries.toList();
    var reduced = <Debt>[];
    for (var combinedDebtEntry in combinedDebtEntries) {
      var debtAmount = combinedDebtEntry.value;
      var creditor = combinedDebtEntry.key;
      var debtor = userId;

      if(debtAmount >= 0) {
        reduced.add(Debt(creditor: creditor, debtor: debtor, amount: debtAmount));
      }
      else {
        reduced.add(Debt(creditor: debtor, debtor: creditor, amount: debtAmount * -1));
      }
    }

    yield* Stream.fromIterable(reduced);
  }
}
