import 'dart:async';

import 'package:dart_appwrite/dart_appwrite.dart';
import 'package:dart_appwrite/models.dart' as appwrite_models;
import 'package:lib/src/exceptions/id_null.dart';
import 'package:lib/src/utils/appwrite_client.dart';
import 'package:lib/src/interface/db_model.dart';

abstract class BaseDatabase<T extends DbModel, T1> extends AppwriteClient {
  static const String itemDataField = "data";

  final Map<String, dynamic> Function(T) itemToDbMap;
  final T1 Function(Map<String, dynamic>) itemFromDbMap;
  
  late final Databases databases;

  String get databaseId;
  String get collectionId;

  BaseDatabase({required super.host, required super.projectId, required super.apiKey, required this.itemFromDbMap, required this.itemToDbMap}) {
    databases = Databases(client);
  }

  Future<T1> create(T item) async {
    var id = item.id ?? ID.unique();
    var document = await databases.createDocument(databaseId: databaseId, collectionId: collectionId, documentId: id, data: itemToDbMap(item), permissions: item.permissions?.dbPermissions);
    return _itemFromDocument(document);
  }

  Future<T1> get(String id) async {
    var document = await databases.getDocument(databaseId: databaseId, collectionId: collectionId, documentId: id);
    return _itemFromDocument(document);
  }

  Stream<T1> getFromQueries(Iterable<String> queries) async* {
    var queryResult = await databases.listDocuments(databaseId: databaseId, collectionId: collectionId, queries: queries.toList());
    if(queryResult.total > 0) {
      for (var document in queryResult.documents) {
        yield _itemFromDocument(document);
      }

      if(!queries.any((query) => query.contains("cursorBefore"))) {
        var cursorQueries = [
          ...queries,
          Query.cursorAfter(queryResult.documents.last.$id),
        ];
        queryResult = await databases.listDocuments(databaseId: databaseId, collectionId: collectionId, queries: cursorQueries);
      }
    }
  }

  Stream<T1> getAll() async* {
    var queryResults = getFromQueries([]);
    await for (var t in queryResults) {
      yield t;
    }
  }

  Future<T1> update(T item) async {
    if (item.id == null) {
      throw IdNullException(item);
    }
    var document = await databases.updateDocument(databaseId: databaseId, collectionId: collectionId, documentId: item.id!, data: itemToDbMap(item), permissions: item.permissions?.dbPermissions);
    return _itemFromDocument(document);
  }

  Future<T1> delete(String id) async {
    var itemToDelete = await get(id);
    await databases.deleteDocument(databaseId: databaseId, collectionId: collectionId, documentId: id);
    return itemToDelete;
  }

  T1 _itemFromDocument(appwrite_models.Document document) {
    return itemFromDbMap(document.toMap()[itemDataField]);
  }
}
