import 'package:lib/src/db_api_endpoint.dart';
import 'package:lib/src/exceptions/id_null.dart';
import 'package:lib/src/exceptions/unauthorized.dart';
import 'package:lib/src/interface/db_model.dart';

class CheckedDbApiEndpoint<T extends DbModel, T1> extends DbApiEndpoint<T, T1> {
  final Future<bool> Function(T)? createItemAllowed;
  final Future<bool> Function(T1)? getItemAllowed;
  final Future<bool> Function(T, T1)? updateItemAllowed;
  final Future<bool> Function(T1)? deleteItemAllowed;
  final bool allowActionsPerDefault;

  CheckedDbApiEndpoint({
    required super.database,
    required super.itemFromRequest,
    required super.itemToResponse,
    this.createItemAllowed,
    this.getItemAllowed,
    this.updateItemAllowed,
    this.deleteItemAllowed,
    this.allowActionsPerDefault = false,
  });

  @override
  Future<T1> create(T item) async {
    await checkAndThrowIfNecessary(createItemAllowed, item);
    return super.create(item);
  }

  @override
  Future<T1> update(T item) async {
    if(item.id == null) {
      throw IdNullException(item);
    }

    var originalItem = await super.getFromId(item.id!);
    await checkAndThrowIfNecessary(updateCheck(item, originalItem), item);

    return super.update(item);
  }

  @override
  Future<T1> delete(String id) async {
    var item = await super.getFromId(id);
    await checkAndThrowIfNecessary(deleteItemAllowed, item);

    return super.delete(id);
  }

  @override
  Future<T1> getFromId(String id) async {
    var item = await super.getFromId(id);
    await checkAndThrowIfNecessary(getItemAllowed, item);

    return item;
  }

  @override
  Stream<T1> getFromQueries(Iterable<String> queries) async* {
    await for (var item in super.getFromQueries(queries)) {
      if(await checkIfAllowed(getItemAllowed, item)) {
        yield item;
      }
    }
  }

  @override
  Stream<T1> getAll() async* {
    await for (var item in super.getAll()) {
      if(await checkIfAllowed(getItemAllowed, item)) {
        yield item;
      }
    }
  }

  Future<void> checkAndThrowIfNecessary<T2>(Future<bool> Function(T2)? check, T2 item) async {
    if(!await checkIfAllowed(check, item)) {
      throw UnauthorizedException();
    }
  }

  Future<bool> checkIfAllowed<T2>(Future<bool> Function(T2)? check, T2 item) {
    if(check != null) {
      return check(item);
    }

    return Future.value(allowActionsPerDefault);
  }

  Future<bool> Function(T)? updateCheck(T item, T1 originalItem) {
    if(updateItemAllowed != null) {
      return (item) => updateItemAllowed!(item, originalItem);
    }
    return null;
  }
}
