import 'dart:async';
import 'dart:convert';

import 'package:lib/src/base_database.dart';
import 'package:lib/src/converter/list_response/list_to_response_converter.dart';
import 'package:lib/src/interface/api_endpoint.dart';
import 'package:lib/src/interface/db_model.dart';
import 'package:lib/src/utils/api_helper.dart';
import 'package:lib/src/utils/variables.dart';

class DbApiEndpoint<T extends DbModel, T1> extends ApiEndpoint {
  final BaseDatabase<T, T1> database;
  final T Function(Map<String, dynamic>) itemFromRequest;
  final Map<String, dynamic> Function(T1) itemToResponse;

  DbApiEndpoint({required this.database, required this.itemFromRequest, required this.itemToResponse});

  @override
  Future<Map<String, dynamic>> handleDelete(context) async {    
    var id = idFromContext(context);
    var item = await delete(id);
    return context.res.json(itemToResponse(item));
  }

  @override
  Future<Map<String, dynamic>> handleGet(context) async {
    if(ApiHelper.hasBodyField(context, RequestResponseVariables.idFieldName)) {
      var id = idFromContext(context);
      var item = await getFromId(id);
      return context.res.json(itemToResponse(item));
    }
    else if(ApiHelper.hasBodyField(context, RequestResponseVariables.queriesFieldName)) {
      var queryRequest = queryRequestFromContext(context);
      var items = await getFromQueries(queryRequest).toList();
      return listResponse(context, items, itemToResponse);
    }
    else {
      var items = await getAll().toList();
      return listResponse(context, items, itemToResponse);
    }
  }

  @override
  Future<Map<String, dynamic>> handlePost(context) async {
    var item = requestFromContext(context);
    var createdItem = await create(item);
    return context.res.json(itemToResponse(createdItem));
  }

  @override
  Future<Map<String, dynamic>> handlePut(context) async {
    var item = requestFromContext(context);
    var updatedItem = await update(item);
    return context.res.json(itemToResponse(updatedItem));
  }

  Future<T1> create(T item) {
    return database.create(item);
  }

  Future<T1> update(T item) {
    return database.update(item);
  }

  Future<T1> delete(String id) {
    return database.delete(id);
  }

  Future<T1> getFromId(String id) {
    return database.get(id);
  }

  Stream<T1> getFromQueries(Iterable<String> queries) {
    return database.getFromQueries(queries);
  }

  Stream<T1> getAll() {
    return database.getAll();
  }

  T requestFromContext(final context) {
    return itemFromRequest(jsonDecode(context.req.body));
  }

  String idFromContext(final context) {
    return ApiHelper.fieldFromRequestBody(context, RequestResponseVariables.idFieldName);
  }

  Iterable<String> queryRequestFromContext(final context) {
    return ApiHelper.fieldFromRequestBody(context, RequestResponseVariables.queriesFieldName);
  }
}
